/* Copyright (C) 2022 Brigham Keys, Esq.
 * This file is part of Zombie Shootoff
 *
 * Zombie Shootoff is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Zombie Shootoff is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with thissoftware.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <iostream>
#include <IrrlichtDevice.h>
#include <nlohmann/json.hpp>

#include <Gwork/Gwork.h>
#include <Gwork/Input/Irrlicht.h>
#include <Gwork/Platform.h>
#include <Gwork/Renderers/Irrlicht.h>
#include <Gwork/Skins/Simple.h>
#include <Gwork/Skins/TexturedBase.h>
#include <Gwork/Test/Test.h>
#include <btBulletDynamicsCommon.h>

#include "Global.hpp"

int main() {
  if (Global::Init::initializeIrrlicht()) {
    return 1;
  }
  int i;
  btDefaultCollisionConfiguration *collisionConfiguration =
      new btDefaultCollisionConfiguration();

  /// use the default collision dispatcher. For parallel processing you can use
  /// a diffent dispatcher (see Extras/BulletMultiThreaded)
  btCollisionDispatcher *dispatcher =
      new btCollisionDispatcher(collisionConfiguration);

  /// btDbvtBroadphase is a good general purpose broadphase. You can also try
  /// out btAxis3Sweep.
  btBroadphaseInterface *overlappingPairCache = new btDbvtBroadphase();

  /// the default constraint solver. For parallel processing you can use a
  /// different solver (see Extras/BulletMultiThreaded)
  btSequentialImpulseConstraintSolver *solver =
      new btSequentialImpulseConstraintSolver;

  btDiscreteDynamicsWorld *dynamicsWorld = new btDiscreteDynamicsWorld(
      dispatcher, overlappingPairCache, solver, collisionConfiguration);

  dynamicsWorld->setGravity(btVector3(0, -10, 0));

  btAlignedObjectArray<btCollisionShape *> collisionShapes;

  /// create a few basic rigid bodies

  // the ground is a cube of side 100 at position y = -56.
  // the sphere will hit it at y = -6, with center at -5
  {
    btCollisionShape *groundShape =
        new btBoxShape(btVector3(btScalar(50.), btScalar(50.), btScalar(50.)));

    collisionShapes.push_back(groundShape);

    btTransform groundTransform;
    groundTransform.setIdentity();
    groundTransform.setOrigin(btVector3(0, -56, 0));

    btScalar mass(0.);

    // rigidbody is dynamic if and only if mass is non zero, otherwise static
    bool isDynamic = (mass != 0.f);

    btVector3 localInertia(0, 0, 0);
    if (isDynamic)
      groundShape->calculateLocalInertia(mass, localInertia);

    // using motionstate is optional, it provides interpolation capabilities,
    // and only synchronizes 'active' objects
    btDefaultMotionState *myMotionState =
        new btDefaultMotionState(groundTransform);
    btRigidBody::btRigidBodyConstructionInfo rbInfo(mass, myMotionState,
                                                    groundShape, localInertia);
    btRigidBody *body = new btRigidBody(rbInfo);

    // add the body to the dynamics world
    dynamicsWorld->addRigidBody(body);
  } // added ground
  irr::scene::IMeshSceneNode *box = Global::smgr->addCubeSceneNode(10.f,
                                                                   0,
                                                                   -1,
                                                                   irr::core::vector3df(0, 5, 0));
  if (box) {
    box->setMaterialFlag(irr::video::EMF_LIGHTING, true);
  }
  {
    // create a dynamic rigidbody

    // btCollisionShape* colShape = new btBoxShape(btVector3(1,1,1));
    btCollisionShape *colShape = new btBoxShape(btVector3(5, 5, 5));
    colShape->setUserPointer(reinterpret_cast<void *>(box));
    collisionShapes.push_back(colShape);

    /// Create Dynamic Objects
    btTransform startTransform;
    startTransform.setIdentity();

    btScalar mass(1.f);

    // rigidbody is dynamic if and only if mass is non zero, otherwise static
    bool isDynamic = (mass != 0.f);

    btVector3 localInertia(0, 0, 0);
    if (isDynamic)
      colShape->calculateLocalInertia(mass, localInertia);

    startTransform.setOrigin(btVector3(2, 40, 0));

    // using motionstate is recommended, it provides interpolation capabilities,
    // and only synchronizes 'active' objects
    btDefaultMotionState *myMotionState =
        new btDefaultMotionState(startTransform);
    btRigidBody::btRigidBodyConstructionInfo rbInfo(mass, myMotionState,
                                                    colShape, localInertia);
    btRigidBody *body = new btRigidBody(rbInfo);

    dynamicsWorld->addRigidBody(body);
  } // create a sphere

  Global::smgr->addCameraSceneNode(0,
                                   irr::core::vector3df(0, 30, -40),
                                   irr::core::vector3df(0, 5, 0));

  while (Global::device->run()) {
    Global::driver->beginScene(true, true, irr::video::SColor(255, 100, 101, 140));
    dynamicsWorld->stepSimulation(1.f / 60.f, 10);

    // print positions of all objects
    for (int j = dynamicsWorld->getNumCollisionObjects() - 1; j >= 0; j--) {
      btCollisionObject *obj = dynamicsWorld->getCollisionObjectArray()[j];
      btRigidBody *body = btRigidBody::upcast(obj);
      btTransform trans;
      if (body && body->getMotionState()) {
          body->getMotionState()->getWorldTransform(trans);
      } else {
          trans = obj->getWorldTransform();
      }
      /*printf("world pos object %d = %f,%f,%f\n",
             j,
             float(trans.getOrigin().getX()),
             float(trans.getOrigin().getY()),
             float(trans.getOrigin().getZ()));*/
      irr::scene::IMeshSceneNode *mesh = reinterpret_cast<irr::scene::IMeshSceneNode *>(
          body->getCollisionShape()->getUserPointer());
      std::cout << Global::device->getTimer()->getSpeed() << std::endl;
      if (mesh not_eq nullptr) {
          mesh->setPosition(irr::core::vector3df(trans.getOrigin().getX(),
                                                 trans.getOrigin().getY(),
                                                 trans.getOrigin().getZ()));
      }
    }
    Global::smgr->drawAll();
    Global::guienv->drawAll();

    Global::driver->endScene();
  }
  Global::device->drop();

  // remove the rigidbodies from the dynamics world and delete them
  for (i = dynamicsWorld->getNumCollisionObjects() - 1; i >= 0; i--) {
    btCollisionObject *obj = dynamicsWorld->getCollisionObjectArray()[i];
    btRigidBody *body = btRigidBody::upcast(obj);
    if (body && body->getMotionState()) {
      delete body->getMotionState();
    }
    dynamicsWorld->removeCollisionObject(obj);
    delete obj;
  }

  // delete collision shapes
  for (int j = 0; j < collisionShapes.size(); j++) {
    btCollisionShape *shape = collisionShapes[j];
    collisionShapes[j] = 0;
    delete shape;
  }

  // delete dynamics world
  delete dynamicsWorld;

  // delete solver
  delete solver;

  // delete broadphase
  delete overlappingPairCache;

  // delete dispatcher
  delete dispatcher;

  delete collisionConfiguration;

  // next line is optional: it will be cleared by the destructor when the array
  // goes out of scope
  collisionShapes.clear();
  return 0;
}
